#https://discourse.panda3d.org/t/using-ssbo-for-light-data/30433/7
#https://discourse.panda3d.org/t/glsl-dithering-filter-help/15010/15
#https://github.com/panda3d/panda3d/issues/630
#https://github.com/ComradeNapCali/libpandabsp/blob/feee2e7d3e102aebf0ed70317d4702f39e324b62/src/libpandabsp/shader_generator.cpp#L153

from panda3d.core import *
load_prc_file_data("", "gl-version 3 1")
load_prc_file_data("", "basic-shaders-only #f")
load_prc_file_data("", "gl-check-errors True gl-debug True gl-debug-object-labels True")
from direct.showbase.ShowBase import ShowBase
from direct.interval.LerpInterval import LerpPosInterval,LerpHprInterval
from direct.interval.IntervalGlobal import *

class Main(ShowBase):

    def __init__(self):

        ShowBase.__init__(self)
        #base.set_background_color(0, 0, 0)

        light = DirectionalLight("MyDirectional")
        light.color = (0.0,0.7,0.3,1.0)
        self._light_np = base.render.attach_new_node(light)
        self._light_np.set_pos(0,-500,10.0)
        self._light_np.node().set_shadow_caster(True,512,512)
        self._light_np.node().get_lens().set_film_size(50.0,50.0)
        self._light_np.node().get_lens().set_near_far(0.1,1000.0)
        base.render.set_light(self._light_np)

        light2 = Spotlight("MySpot")
        light2.color = (5.0,0.1,0.0,1.0)
        light2.max_distance = 8000.0
        self._light2_np = base.render.attach_new_node(light2)
        self._light2_np.set_pos(0,0,25)
        self._light2_np.node().set_shadow_caster(True,512,512)
        base.render.set_light(self._light2_np)

        np = base.loader.load_model(Filename.from_os_specific("box"))
        np.reparent_to(base.render)
        np.set_two_sided(False)

        np2 = base.loader.load_model(Filename.from_os_specific("box"))
        np2.reparent_to(base.render)
        np2.set_two_sided(False)
        np2.set_pos(0.0,-10.0,-30.0)
        np2.set_scale(15.0)

        np.set_shader_auto(True)
        np2.set_shader_auto(True)

        self._light2_np.look_at(np)
        self._light_np.look_at(np)

        self._tex_array = Texture("volume")
        self._tex_array.setup2dTextureArray(512, 512, 2, Texture.T_float, Texture.F_depth_component32)
        self._tex_array.set_clear_color( LVecBase4( 1.0 ))
        self._state = SamplerState()
        self._state.setMinfilter( SamplerState.FT_linear)
        self._state.setMagfilter( SamplerState.FT_linear )
        self._state.set_wrap_u( SamplerState.WM_clamp )
        self._state.set_wrap_v( SamplerState.WM_clamp )
        self._tex_array.setDefaultSampler( self._state )

        self._tex_array2 = Texture("volume2")
        self._tex_array2.setup2dTextureArray(512, 512, 1, Texture.T_float, Texture.F_depth_component32)
        self._tex_array2.set_clear_color( LVecBase4( 1.0 ))
        self._state = SamplerState()
        self._state.setMinfilter( SamplerState.FT_linear)
        self._state.setMagfilter( SamplerState.FT_linear )
        self._state.set_wrap_u( SamplerState.WM_clamp )
        self._state.set_wrap_v( SamplerState.WM_clamp )
        self._tex_array2.setDefaultSampler( self._state )

        self._start_pos = self._light2_np.get_pos()
        self._end_pos = Point3(Vec3(self._start_pos.get_x()-10.0,self._start_pos.get_y(),self._start_pos.get_z()+50.0) ) 
        interval1 = LerpPosInterval(self._light2_np,5.0,pos = self._end_pos,startPos = self._start_pos) 
        interval2 = LerpPosInterval(self._light2_np,5.0,pos = self._start_pos)
        self._seq = Sequence(interval1, interval2,name="MySequence")
        self._seq.loop()

        base.taskMgr.add(self._capture)


    def _capture(self,task):
        sBuffer = self._light_np.node().getShadowBuffer(GraphicsStateGuardianBase.getGsg(0))
        sBuffer2 = self._light2_np.node().getShadowBuffer(GraphicsStateGuardianBase.getGsg(0))
        if sBuffer is not None and sBuffer2 is not None:
            self.cardmaker = CardMaker("cubemaker")
            self.cardmaker.setFrame(-1,1,-1,1)
            for layer in range(2): 
                self.cardmaker.setUvRange((0, 1, 1, 0), (0, 0, 1, 1),(layer, layer, layer, layer))
                card = NodePath(self.cardmaker.generate())
                card.setTexture(self._tex_array)
                card.set_two_sided(True)
                card.set_scale(0.3)
                card.set_pos(-0.5 + 0.8*layer,0.0,0.6)
                shader = self._generate_shader(False,None)
                card.set_shader(shader,1000)
                if layer ==0:
                    card.set_shader_input("texture_array",self._tex_array)
                else:
                    card.set_shader_input("texture_array",self._tex_array2)
                card.set_shader_input("DMKlayer",layer,10)
                card.reparent_to(base.render2d)
            #Light1
            sBuffer.addRenderTexture(self._tex_array, GraphicsOutput.RTM_bind_layered, GraphicsOutput.RTP_depth) 
            sBuffer.get_display_region(0).set_clear_depth_active(True)
            attr = self._generate_shader(True,0)
            state = self._light_np.node().get_initial_state()
            state = state.add_attrib(attr, 10)
            self._light_np.node().set_initial_state(state)
            ####Light2: change self._tex_array2 below by self._tex_array: this will not work anymore, I can't use one sliced texture...
            sBuffer2.addRenderTexture(self._tex_array2, GraphicsOutput.RTM_bind_layered, GraphicsOutput.RTP_depth)
            sBuffer2.get_display_region(0).set_clear_depth_active(True)
            attr = self._generate_shader(True,1)
            state = self._light2_np.node().get_initial_state()
            state = state.add_attrib(attr, 10)
            self._light2_np.node().set_initial_state(state)
            return task.done
        return task.cont

    def _generate_shader(self,is_geom,layer):
        #Set the GLSL Target to the shader
        file = open("vertex_VS.glsl")
        vs = file.read()
        file = open("fragment_FS.glsl")
        fs = file.read()
        file = open("geom_GS.glsl")
        geom = file.read()
        if is_geom is False:
            return Shader.make(Shader.SL_GLSL, vertex = vs, fragment = fs)
        else:
            file = open("PassThrough_FS.glsl")
            fs = file.read()
            shader =  Shader.make(Shader.SL_GLSL, vertex = vs, fragment = fs, geometry = geom)
            attr = ShaderAttrib.make(shader)
            attr = attr.set_shader_input("Mylayer",layer,1000)
            print(attr.get_shader_input("Mylayer").get_vector())
            return attr


a = Main()
a.run()

